import os.path as path
import argparse
import re
from . import __version__


def validate_multijaccard(ARGS):
    """
    Validate command line arguments for multijaccard

    Parameters
    ----------
    ARGS : Namespace
        Command line arguments
    """
    # check for files existing
    nonex_files = [b for b in ARGS.bed if not path.exists(b)]
    if len(nonex_files) != 0:
        raise OSError(" ".join([nonex_files[0], "not found."]))
    # check there are the same number of names as BED files
    if ARGS.names is not None:
        names = ARGS.names.split(",")
        if len(names) != len(ARGS.bed):
            raise ValueError("`names` and `beds` parameters must be the same length")
    else:
        names = None
    # check output file formats for plots
    if ARGS.exts is not None:
        exts = ARGS.exts.split(",")
    return {
        "beds": ARGS.bed,
        "names": names,
        "prefix": ARGS.prefix,
        "plot": not ARGS.no_plot,
        "exts": exts,
    }


def validate_fastq_info(ARGS):
    """
    Validate command line arguments for multijaccard

    Parameters
    ----------
    ARGS : Namespace
        Command line arguments
    """
    # check file exists
    if not path.exists(ARGS.fastq):
        raise OSError("`{}` not found.".format(ARGS.fastq))
    return {"fastq": ARGS.fastq}


def validate_filter_qname(ARGS):
    """
    Validate command line arguments for filter-qname

    Parameters
    ----------
    ARGS : Namespace
        Command line arguments
    """
    # check file exists
    if not path.exists(ARGS.fastq):
        raise OSError("`{}` not found.".format(ARGS.fastq))
    return {"bam": ARGS.fastq}


def validate_organize(ARGS):
    """
    Validate command line arguments for organize

    Parameters
    ----------
    ARGS : Namespace
        Command line arguments
    """
    from .data import DIRNAME_REGEX

    # check folder exists
    if not path.exists(ARGS.dir):
        raise OSError("`{}` not found.".format(ARGS.dir))
    # check directory path matches expected regex for raw sequencing data
    # matches YYMMDD_InstrumentSerialNumber_RunNumber_(A|B)FlowcellID
    pattern = re.compile(DIRNAME_REGEX)
    if not pattern.fullmatch(ARGS.dir):
        raise OSError(
            "`{}` does not match the expected folder name format.".format(ARGS.dir),
            "Should match `YYMMDD_InstrumentSerialNumber_RunNumber_(A|B)FlowcellID[_OPTIONAL_TEXT]`.",
        )
    return {"indir": ARGS.dir, "outdir": ARGS.outdir, "seqtype": ARGS.type}


def main():
    PARSER = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    # global options
    PARSER.add_argument(
        "--version", action="store_true", help="Print version number and exit"
    )

    SUBPARSERS = PARSER.add_subparsers(
        title="Sub-commands", dest="command", metavar="<command>"
    )

    # multi-jaccard
    multijaccard_parser = SUBPARSERS.add_parser(
        "multi-jaccard",
        help="Perform `bedtools jaccard` on multiple BED files, and cluster samples by their pair-wise Jaccard indices.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    multijaccard_parser.add_argument("bed", type=str, help="BED file(s)", nargs="+")
    multijaccard_parser.add_argument(
        "-n", "--names", type=str, help="Comma-separated labels for BED files"
    )
    multijaccard_parser.add_argument(
        "-o", "--prefix", type=str, help="Prefix for output files.", default="jaccard"
    )
    multijaccard_parser.add_argument(
        "-p",
        "--no-plot",
        action="store_true",
        help="Don't plot, only produce table",
        default=False,
    )
    multijaccard_parser.add_argument(
        "-e",
        "--exts",
        type=str,
        help="Comma-separated list of image file extensions to plot",
        default="png",
    )

    # fastq-info
    fastqinfo_parser = SUBPARSERS.add_parser(
        "fastq-info",
        help="Extract metadata from read information in a FASTQ.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    fastqinfo_parser.add_argument("fastq", type=str, help="BED file(s)")

    # filter-qname
    filter_qname_parser = SUBPARSERS.add_parser(
        "filter-qname", help="Filter a SAM/BAM file by its query names"
    )
    filter_qname_parser.add_argument(
        "bam", type=str, help="Query name-sorted BAM file to be filtered"
    )
    filter_qname_parser.add_argument(
        "ids", type=str, help="Text file containing IDs to be removed"
    )
    filter_qname_parser.add_argument(
        "-o", "--output", type=str, help="Output file", default="filtered.bam"
    )

    # org
    org_parser = SUBPARSERS.add_parser(
        "org",
        help="Organize a batch of raw sequencing data",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    org_parser.add_argument("dir", type=str, help="Input directory to organize")
    org_parser.add_argument(
        "-o", "--outdir", type=str, help="New path for input directory"
    )
    org_parser.add_argument(
        "-t",
        "--type",
        type=str,
        choices=["atac", "dname", "chip", "dna", "rna", "hic", "mix"],
        help="Type of sequencing data in batch.",
        default="mix",
    )

    # parse arguments from command line
    ARGS = PARSER.parse_args()

    # validate command line arguments for the give sub-command
    # import packages after parsing to speed up command line responsiveness
    if ARGS.version:
        print("bio-jtools v{}".format(__version__))
        return

    if ARGS.command == "multi-jaccard":
        validated_args = validate_multijaccard(ARGS)
        from .interval.multijaccard import multijaccard

        func = multijaccard
    elif ARGS.command == "fastq-info":
        validated_args = validate_fastq_info(ARGS)
        from .fastx.fastq_info import fastq_info

        func = fastq_info
    elif ARGS.command == "filter-qname":
        validated_args = validate_filter_qname(ARGS)
        from .align.filter_qname import filter_qname

        func = filter_qname
    elif ARGS.command == "org":
        validated_args = validate_organize(ARGS)
        from .data.organize import organize

        func = organize
    else:
        pass
    # using the func = ... format allows for a single universal function call
    val = func(**validated_args)
    if val is not None:
        print(val)



from __future__ import division, absolute_import, print_function
import re
from datetime import datetime
from os import listdir, mkdir, rename
import os.path as path
from . import DIRNAME_REGEX, FASTQ_FILENAME_REGEX, RESERVED_DIRS, RESERVED_FILENAMES, README_STR, CLUSTER_STR, SETUP_STR, SNAKEFILE_STR


def fetch_seq_info(dirname):
    '''
    Extract sequencing batch information from input directory filename

    Parameters
    ----------
    dirname : str
        Directory filename to parse
    '''
    # check directory path matches expected regex for raw sequencing data
    # matches YYMMDD_InstrumentSerialNumber_RunNumber_(A|B)FlowcellID[_OPTIONAL_TEXT]
    # see https://www.biostars.org/p/124972/, https://www.biostars.org/p/198143/
    pattern = re.compile(DIRNAME_REGEX)
    m = re.match(pattern, dirname)
    vals = {
        'date': datetime.strptime(''.join([m.group(1), m.group(2), m.group(3)]), '%y%m%d'),
        'instrument': m.group(4),
        'run': m.group(5),
        'position': m.group(6),
        'flowcell': m.group(7),
        'date_sub': '',
        'description': m.group(8)
    }
    vals['date_rec'] = vals['date'].strftime('%Y-%m-%d')
    return vals


def create_cluster_params(outfile):
    '''
    Create `cluster.yaml` file

    Parameters
    ----------
    outfile : str
        Output config file to create
    '''
    fh = open(outfile, 'w')
    fh.write(CLUSTER_STR)
    fh.close()


def create_readme(seq_info, outfile):
    '''
    Create `README.md` file

    Parameters
    ----------
    seq_info : Dict of {str, Object}
        Sequencing metadata retrieved from directory name
    outfile : str
        Output config file to create
    '''
    fh = open(outfile, 'w')
    fh.write(README_STR.format(**seq_info))
    fh.close()


def create_config(fastq_dir, outfile):
    '''
    Create `config.tsv` file

    Parameters
    ----------
    fastq_dir : str
        Directory containing FASTQs files to parse
    outfile : str
        Output config file to create
    '''
    import pandas as pd
    # record renaming actions
    actions = []
    # get all FASTQs in directory
    fastqs = listdir(fastq_dir)
    # create placholder for tracking samples and their respective files
    samples_dict = {}
    # check that file names match FASTQ regex
    pattern = re.compile(FASTQ_FILENAME_REGEX)
    for f in fastqs:
        # extract sample information from file name
        m = pattern.match(f)
        if m is None:
            continue
        name = m.group(1)
        index = m.group(2)
        lane = m.group(3)
        mate = m.group(4)
        zipped = m.group(5)
        cname = correct_sample_name(name)
        if cname not in samples_dict:
            samples_dict[cname] = {'sid': index, 'lanes': set([lane]), 'mates': set([mate])}
        else:
            if index != samples_dict[cname]['sid']:
                raise IOError('`{}` has multiple indexes on this same flowcell: {}'.format(
                    f, ' '.join(index, samples_dict[cname]['sid'])))
            else:
                samples_dict[cname]['lanes'].add(lane)
                samples_dict[cname]['mates'].add(mate)
        # create new file name for FASTQ
        # remove sample indexes and `_001` from file names
        newf = '_'.join([cname, 'L00' + lane, mate]) + '.fastq' + zipped
        src = path.join(fastq_dir, f)
        dest = path.join(fastq_dir, newf)
        # rename files if they don't abide by the format
        if src != dest:
            rename(src, dest)
            actions.append(' '.join([src, '->', dest]))
    # fill in sample configuration table for config.tsv
    samples_df = pd.DataFrame(columns=['Sample', 'Sample_Index_Number', 'Lanes', 'Mates'])
    for k, v in samples_dict.items():
        samples_df = samples_df.append(
            {
                'Sample': k,
                'Sample_Index_Number': v['sid'],
                'Lanes': ','.join(v['lanes']),
                'Mates': ','.join(v['mates'])
            },
            ignore_index=True
        )
    # write table to outfile
    samples_df.to_csv(outfile, sep='\t', index=False)
    return actions


def correct_sample_name(name):
    '''
    Correct FASTQ's sample name to match the following regex: '[A-Za-z0-9-]+'.

    Parameters
    ----------
    name : str
        Filename to be corrected
    '''
    newname = name.replace('_', '-')
    return newname


def create_snakefile(seqtype, outfile):
    '''
    Create `Snakefile` config file

    Parameters
    ----------
    seqtype : str
        Type of sequencing data contained in the input folder.
        Must be one of ['atac', 'dname', 'chip', 'dna', 'rna', 'hic', 'mix'].
    outfile : str
        Output config file to create
    '''
    # write to Snakefile
    fh = open(outfile, 'w')
    fh.write(SNAKEFILE_STR)
    fh.close()


def organize(indir, outdir=None, seqtype='mix'):
    '''
    Organize raw sequencing data folder

    Parameters
    ----------
    indir : str
        Input directory to organize
    outdir : str
        New path for input directory
    seqtype : str
        Type of sequencing data contained in the input folder.
        Must be one of ['atac', 'dname', 'chip', 'dna', 'rna', 'hic', 'mix'].
    '''
    # record steps performed during setup
    setup_log = {'mkdir': [], 'mv': [], 'res_files': [], 'fastq': []}
    # extract sequencing metadata information from directory name
    seq_info = fetch_seq_info(indir)
    default_dirs = RESERVED_DIRS
    if seqtype == 'atac' or seqtype == 'chip':
        default_dirs += ['Peaks']
    elif seqtype == 'dname':
        default_dirs += ['Methylation']
    elif seqtype == 'hic':
        default_dirs += ['Contacts']
    # make directories
    for d in default_dirs:
        dir_to_make = path.join(indir, d)
        if not path.exists(dir_to_make):
            mkdir(dir_to_make)
            # add action to setup log
            setup_log['mkdir'].append(d)
    # find FASTQs in directory
    dir_files = listdir(indir)
    fastq_files = [f for f in dir_files if f.endswith('.fastq.gz')]
    other_files = [f for f in dir_files if not f.endswith('.fastq.gz') and f not in RESERVED_FILENAMES]
    # move FASTQs into `FASTQs` directory
    for f in fastq_files:
        src = path.join(indir, f)
        dest = path.join(indir, 'FASTQs', f)
        rename(src, dest)
        # add action to setup log
        setup_log['mv'].append(' '.join([src, '->', dest]))
    for f in other_files:
        src = path.join(indir, f)
        dest = path.join(indir, 'Reports', f)
        if not path.isdir(src):
            rename(src, dest)
            # add action to setup log
            setup_log['mv'].append(' '.join([src, '->', dest]))
        else:
            if f not in RESERVED_DIRS:
                rename(src, dest)
                # add action to setup log
                setup_log['mv'].append(' '.join([src, '->', dest]))
    # check for README and other config files in directory
    reserved_files = [path.join(indir, f) for f in RESERVED_FILENAMES]
    if not path.exists(reserved_files[0]):
        create_readme(seq_info, reserved_files[0])
        # add action to setup log
        setup_log['res_files'].append(reserved_files[0])
    if not path.exists(reserved_files[1]):
        create_cluster_params(reserved_files[1])
        # add action to setup log
        setup_log['res_files'].append(reserved_files[1])
    # create config and add all renaming events to setup log
    setup_log['fastq'] = create_config(path.join(indir, RESERVED_DIRS[1]), reserved_files[2])
    # add creation of config file to setup log
    setup_log['res_files'].append(reserved_files[2])
    if not path.exists(reserved_files[3]):
        create_snakefile(seqtype, reserved_files[3])
        # add action to setup log
        setup_log['res_files'].append(reserved_files[3])
    # reformat setup_log for printing
    setup_log_printable = {}
    for k, v in setup_log.items():
        if len(v) > 0:
            setup_log_printable[k] = '\n'.join(setup_log[k])
        else:
            setup_log_printable[k] = ''
    fh = open(path.join(indir, 'setup.log'), 'w')
    fh.write(SETUP_STR.format(**setup_log_printable))
    fh.close()
    # rename directory if asked to
    if outdir is not None:
        rename(indir, outdir)



from Bio import SeqIO
from gzip import open as gzopen
from itertools import islice, zip_longest
import sys

def grouper(iterable, n, fillvalue=None):
    '''
    Collect data into fixed-length chunks or blocks

    Parameters
    ----------
    iterable : 
        Path to FASTQ file
    '''
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return zip_longest(fillvalue=fillvalue, *args)


def parse_read_id(read_id, tag=None):
    '''
    Extract metadata from a read ID

    Parameters
    ----------
    read_id : str
        Read ID
    tag : str
        Tag to extract from metadata
    '''
    split_info = read_id.split(':')
    tags = ['instrument', 'run', 'flowcell', 'lane', 'tile', 'x', 'y']
    meta = dict(zip(tags, split_info))
    if tag is not None and tag in tags:
        return meta[tag]
    else:
        return meta
    

def fastq_info(fastq, chunksize=10000):
    '''
    Extract flowcell and other metadata from a FASTQ

    Parameters
    ----------
    fastq : str
        Path to FASTQ file
    chunksize : int
        Number of records to read simultaneously
    '''
    # check for file format
    if fastq.endswith('.gz'):
        fq_handle = gzopen(fastq, 'rt')
    else:
        fq_handle = open(fastq, 'r')
    # load reads for random access
    records = SeqIO.parse(fq_handle, 'fastq')
    # data structure for managing metadata
    metadata = {
        'flowcells': set([]),
        'n_reads': 0
    }
    # load `chunksize` reads at a time
    chunked_records = grouper(records, chunksize)
    for chunk in chunked_records:
        # filter chunk
        chunk = [r for r in chunk if r is not None]
        # count reads
        metadata['n_reads'] += len(chunk)
        # extract unique flowcell IDs
        flowcells = set([parse_read_id(r.id, 'flowcell') for r in chunk])
        for f in flowcells:
            metadata['flowcells'].add(f)

    fq_handle.close()
    return metadata




def detect_filetype_from_path(path):
    '''
    Determine what type of file a given path is based on its extension

    Parameters
    ----------
    path : str
        Path to check
    '''
    # convert to lowercase for simpler comparisons
    path = path.tolower()
    # remove '.gz' at the end of a pathname if it is present
    if path.endswith('.gz'):
        path = path[:-4]
    # now parse other extensions
    if path.endswith('.sam'):
        return 'SAM'
    elif path.endswith('.bam'):
        return 'BAM'
    elif path.endswith('.cram'):
        return 'CRAM'
    elif path.endswith('.fasta'):
        return 'FASTA'
    elif path.endswith('.fastq'):
        return 'FASTQ'
    elif path.endswith('.vcf'):
        return 'VCF'
    elif path.endswith('.bcf'):
        return 'BCF'
    elif path.endswith('.maf'):
        return 'MAF'
    elif path.endswith('.tbx'):
        return 'TABIX'
    elif path.endswith('.gtf'):
        return 'TABIX'
    elif path.endswith('.gff'):
        return 'TABIX'
    elif path.endswith('.bed'):
        return 'BED'
    elif path.endswith('.bedpe'):
        return 'BEDPE'
    else:
        return None


def validate_filetype(path):
    '''
    Perform a series of validations on your file of interest

    Parameters
    ----------
    path : str
        File path to validate
    '''
    ftype = detect_filetype_from_path(path)
    func_switch = {
        'SAM': validate_sam,
        'BAM': validate_bam,
        'CRAM': validate_cram,
        'FASTA': validate_fasta,
        'FASTQ': validate_fastq,
        'VCF': validate_vcf,
        'BCF': validate_bcf,
        'MAF': validate_maf,
        'TABIX': validate_tabix,
        'BED': validate_bed,
        'BEDPE': validate_bedpe
    }
    return func_switch[ftype](path)


def validate_sam(path):
    '''
    Validate SAM file

    Parameters
    ----------
    path : str
        File to check
    '''
    raise NotImplementedError()


def validate_bam(path):
    '''
    Validate BAM file

    Parameters
    ----------
    path : str
        File to check
    '''
    raise NotImplementedError()


def validate_cram(path):
    '''
    Validate CRAM file

    Parameters
    ----------
    path : str
        File to check
    '''
    raise NotImplementedError()


def validate_fasta(path):
    '''
    Validate FASTA file

    Parameters
    ----------
    path : str
        File to check
    '''
    raise NotImplementedError()


def validate_fastq(path):
    '''
    Validate FASTQ file

    Parameters
    ----------
    path : str
        File to check
    '''
    raise NotImplementedError()


def validate_vcf(path):
    '''
    Validate VCF file

    Parameters
    ----------
    path : str
        File to check
    '''
    raise NotImplementedError()


def validate_bcf(path):
    '''
    Validate BCF file

    Parameters
    ----------
    path : str
        File to check
    '''
    raise NotImplementedError()


def validate_maf(path):
    '''
    Validate MAF file

    Parameters
    ----------
    path : str
        File to check
    '''
    raise NotImplementedError()


def validate_tabix(path):
    '''
    Validate TABIX file

    Parameters
    ----------
    path : str
        File to check
    '''
    raise NotImplementedError()


def validate_bed(path):
    '''
    Validate BED file

    Parameters
    ----------
    path : str
        File to check
    '''
    raise NotImplementedError()


def validate_bedpe(path):
    '''
    Validate BEDPE file

    Parameters
    ----------
    path : str
        File to check
    '''
    raise NotImplementedError()



from tqdm import tqdm
import seaborn as sns
import matplotlib.pyplot
import numpy as np
import pandas as pd
import pybedtools as pbt

import matplotlib

# handle matplot lib backend nonsense
matplotlib.use("agg")


def multijaccard(beds, names=None, prefix="jaccard", plot=True, exts=["png"]):
    """
    Run bedtools jaccard for pairs of samples and plot as heatmap

    Parameters
    ==========
    beds : list<str>
        BED file paths
    names : list<str>
        Labels for BED files instead of generic 1:N. Assumed to be same length
        as `beds`.
    prefix : str
        Prefix for output files
    plot : bool
        Whether to produce a plot or not
    exts : list<str>
        Output file formats for heatmap image
    """
    # placeholder for resultant data
    results = pd.DataFrame(
        columns=[
            "Label 1",
            "Label 2",
            "Sample 1",
            "Sample 2",
            "intersection",
            "union-intersection",
            "jaccard",
            "n_intersections",
        ]
    )
    for i, a in enumerate(tqdm(beds)):
        # initialize res dictionary
        if names is not None:
            res = {"Sample 1": a, "Label 1": names[i]}
        else:
            res = {"Sample 1": a, "Label 1": int(i)}
        bedA = pbt.BedTool(a)
        for j, b in enumerate(beds[i:]):
            bedB = pbt.BedTool(b)
            # add second sample file
            res["Sample 2"] = b
            if names is not None:
                res["Label 2"] = names[i + j]
            else:
                res["Label 2"] = int(i + j)
            # add dictionary result from bedtools jaccard to res
            res.update(bedA.jaccard(bedB))
            # append this as a new record in the DataFrame
            results.loc[len(results)] = res
    # save output table
    results.to_csv(prefix + ".tsv", sep="\t", index=False)
    if plot:
        # pivot for heatmap
        jaccard = results.pivot("Label 2", "Label 1", "jaccard")
        # mask for clean plotting
        mask = np.zeros_like(jaccard)
        mask[np.triu_indices_from(mask, 1)] = True
        g = sns.heatmap(
            jaccard,
            annot=True,
            fmt=".2f",
            mask=mask,
            square=True,
            annot_kws={"size": 8},
        )
        for e in tqdm(exts):
            g.get_figure().savefig(prefix + "." + e, figsize=(2 * len(beds)))




DIRNAME_REGEX = '^([0-9]{2})(0?[1-9]|1[012])(0[1-9]|[12]\\d|3[01])_(\\w{6})_(\\d{4})_(A|B)(\\w{9})(.*)?/?$'
FASTQ_FILENAME_REGEX = '^([A-Za-z0-9-_]+)_S([1-9][0-9]?)_L00(\\d)_(I1|R[1-3])_001\\.fastq(\\.gz)?$'
RESERVED_DIRS = ['Reports', 'FASTQs', 'Trimmed', 'Aligned']
RESERVED_FILENAMES = ['README.md', 'cluster.yaml', 'config.tsv', 'Snakefile', 'setup.log']

README_STR = '''# Summary

Date Submitted: {date_sub}
Date Received: {date_rec}
Flowcell ID: {flowcell}

{description}
'''

CLUSTER_STR = '''__default__:
  params: "-q lupiengroup -cwd -V"

'''

SNAKEFILE_STR = '''import pandas as pd
import os.path as path

# =============================================================================
# Configuration
# =============================================================================
CONFIG = pd.read_csv('config.tsv', index_col=False, sep='\\t')
CONFIG = CONFIG.loc[~CONFIG.Sample.str.startswith('#'), :]

REPORT_DIR = 'Reports'
FASTQ_DIR = 'FASTQs'
ALIGN_DIR = 'Aligned'

SAMPLES = CONFIG['Sample'].tolist()
READS = [1, 2]
LANES = [1, 2, 3, 4]
BWT2_IDX = '/cluster/tools/data/genomes/human/hg38/iGenomes/Sequence/Bowtie2Index/genome'
CHRS = ['chr' + str(i) for i in list(range(1, 23)) + ['X', 'Y']]

wildcard_constraints:
    sample = '[A-Za-z0-9-]+',
    lane = '[1-4]',
    read = '[1-2]'

# =============================================================================
# Meta Rules
# =============================================================================
rule all:
    input:
        'rulegraph.png',
        path.join(REPORT_DIR, 'multiqc_report.html'),
        expand(
            path.join(REPORT_DIR, '{{sample}}_L00{{lane}}_R{{read}}_fastqc.zip'),
            sample=SAMPLES,
            lane=LANES,
            read=READS
        ),

rule rulegraph:
    output:
        'rulegraph.png',
    shell:
        'snakemake --rulegraph | dot -Tpng > {{output}}'

# =============================================================================
# Rules
# =============================================================================
# Summaries
# -----------------------------------------------------------------------------
rule fastqc:
    input:
        path.join(FASTQ_DIR, '{{file}}.fastq.gz')
    output:
        path.join(REPORT_DIR, '{{file}}_fastqc.html'),
        path.join(REPORT_DIR, '{{file}}_fastqc.zip')
    params:
        '-o {{}}'.format(REPORT_DIR)
    shell:
        'fastqc {{params}} {{input}}'


rule multiqc:
    input:
        samples = expand(
            path.join(REPORT_DIR, '{{sample}}_fastqc.zip'),
            sample=SAMPLES
        )
    output:
        path.join(REPORT_DIR, 'multiqc_report.html')
    shell:
        'multiqc -f -o {{REPORT_DIR}} {{REPORT_DIR}}'



# Miscellaneous
# -----------------------------------------------------------------------------
rule sort_bam_name:
    input:
        '{{file}}.bam'
    output:
        '{{file}}.name-sorted.bam',
    shell:
        'sambamba sort -t 8 --tmpdir . -n -p -o {{output}} {{input}}'

rule sort_bam:
    input:
        '{{file}}.bam'
    output:
        bam = '{{file}}.sorted.bam',
        idx = '{{file}}.sorted.bam.bai'
    shell:
        'sambamba sort -t 8 --tmpdir . -p {{input}}'

'''

SETUP_STR = '''# Making directories
{mkdir}

# Moving files
{mv}

# Creating reserved files
{res_files}

# Formatting FASTQ filenames
{fastq}

'''


from __future__ import division, absolute_import, print_function
import pysam
from tqdm import tqdm
from ..utils import detect_filetype_from_path


def samprint(alignment, output=None):
    '''
    Print to output file or STDOUT if none is supplied

    Parameters
    ----------
    alignment : pysam.AlignedSegment
        Alignment to be written
    outfile : str
        Output file
    '''
    if output is None:
        print(alignment)
    else:
        output.write(alignment)


def filter_qname(bamfile, idfile, outfile=None):
    '''
    Filter reads in a SAM/BAM file by their query names
    '''
    # read in name-sorted BAM file
    bam = pysam.AlignmentFile(bamfile, "rb")
    # output BAM file
    ftype = detect_filetype_from_path(outfile)
    if ftype is None:
        output = None
    elif ftype == 'SAM':
        output = pysam.AlignmentFile(outfile, "w", template=bam)
    elif ftype == 'BAM':
        output = pysam.AlignmentFile(outfile, "wb", template=bam)
    else:
        raise ValueError("Unknown output file format for `outfile`: {}".format(outfile))
    #  read in IDs to be removed (list(set(...)) to only keep unique IDs)
    ids = list(set([l.rstrip() for l in open(idfile, 'r').readlines()]))
    # sort IDs to match BAM for efficient processing (destructive function)
    ids.sort(key=str.lower, reverse=True)
    n_ids = len(ids)
    # progress bar using total alignment counts
    pbar = tqdm()
    reads = bam.fetch(until_eof=True)
    read = next(reads)
    # variable for ensuring BAM is sorted by query name
    last_q = reads.query_name
    while True:
        if read.query_name < last_q:
            raise ValueError(
                'Alignment file is not sorted. {} < {}'.format(read.query_name, last_q)
            )
        # if read name is greater than current top of stack
        if read.query_name > ids[-1]:
            # if this is the last ID
            if n_ids == 1:
                # write remaining reads to new file and break out of while loop
                samprint(read, output)
                for r in bam:
                    samprint(read, output)
                break
            # otherwise pop that ID, try again
            else:
                ids.pop()
                n_ids -= 1
        # skip reads that match the top of the stack
        elif read.query_name == ids[-1]:
            read = next(reads)
        # if read name is less that top of stack, write it and move on
        else:
            samprint(read, output)
            last_q = read.query_name
            read = next(reads)
        pbar.update()
    if output is not None:
        output.close()



from setuptools import setup, find_packages
from os import path

this_dir = path.abspath(path.dirname(__file__))
with open(path.join(this_dir, 'README.md')) as f:
    long_description = f.read()

# version is defined in jtools/__init__.py
__version__ = 'Undefined'
for line in open(path.join('jtools', '__init__.py')):
    if (line.startswith('__version__')):
        exec(line.strip())

setup(
    name='bio-jtools',
    version=__version__,
    description='Various bioinformatics tools in one package',
    url='http://github.com/jrhawley/bio-jtools',
    author='James Hawley',
    author_email='james.hawley@mail.utoronto.ca',
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Natural Language :: English'
    ],
    packages=find_packages(),
    install_requires=[
        'numpy>=1.11',
        'pandas >= 0.15.0',
        'pybedtools',
        'biopython',
        'seaborn',
        'tqdm'
    ],
    entry_points={
        'console_scripts': [
            'jtools=jtools._cli:main'
        ]
    },
    zip_safe=True,
    long_description=long_description,
    long_description_content_type='text/markdown'
)
